# utf-8
"""
Tool Name: Water Extraction Automatically from the image
Source Name: WaterExtraction
Author: Dandan Xu, Nanjing Forestry University
This tool perform Local Moran's I for Raster format, and then extract water body from Local Moran's I result
"""
### Imports 
import arcpy 
from arcpy import env
from arcpy.sa import *
import numpy 
import arcgisscripting
import os
import math
import scipy.stats as st

#set input and output parameters 
inRaster = arcpy.GetParameterAsText(0)
clip_feature = arcpy.GetParameterAsText(1)

###LMIRaster = arcpy.GetParameterAsText(3)
###LMZRaster = arcpy.GetParameterAsText(4)
###LMPRaster = arcpy.GetParameterAsText(5)

Lake = arcpy.GetParameterAsText(2)
#define the temporary file location
tempfiles = arcpy.GetParameterAsText(3)
if tempfiles == '#' or not tempfiles:
    tempfiles = "D:\\waterextraction\\tempfiles" # provide a default value if unspecified

arcpy.CheckOutExtension("Spatial")
gp = arcgisscripting.create(10.5)
gp.OverwriteOutput = 1
gp.workspace = tempfiles
gp.scratchworkspace = tempfiles

# ----------------------------------------------------------------------------
#clip the input raster and calculate the B7 - B4 to control the mountain shadow

inRas = ExtractByMask(inRaster, clip_feature)
myRaster = inRas * 0.0001
myRaster.save("myRaster")

###B4clip = ExtractByMask(B4, clip_feature)
###B4Ras = B4clip * 0.0001
###B7B4 = (myRaster - B4Ras) / (myRaster + B4Ras)
###B7B4Sel = Con((B7B4 < - 0.2), 1 ,0)

# Set environmental variables for output
arcpy.env.outputCoordinateSystem = inRaster 
arcpy.env.cellSize = inRaster
lowerLeft = arcpy.Point(myRaster.extent.XMin, myRaster.extent.YMin)
cellSize = myRaster.meanCellWidth

# Convert Raster to numpy array; calculate based statistics for the array and setup weights (1/8)
# "n" means "the total number of pixels for the raster"
# "nrow" means total number of rows
# "ncol" means total number of columns
# 'x_w' means the row standardized weight 
my_array = arcpy.RasterToNumPyArray(myRaster, nodata_to_value = 0)
my_array.astype(float)
nrow = my_array.shape[0]
ncol = my_array.shape[1]
# calculate mean and std value but not include the background nodata (now equals to 0)
n = 0
x_sum = 0
x_sum2 = 0
for i in range(0, nrow):
    for j in range(0, ncol):
        if my_array[i,j] != 0:
            x_sum = x_sum + float(my_array[i,j])
            n = float(n) + 1
x_ave = x_sum / n
for i in range(0, nrow):
    for j in range(0, ncol):
        if my_array[i,j] != 0:
            x_sum2 = x_sum2 + math.pow((float(my_array[i,j]) - x_ave), 2)
x_std = math.sqrt(x_sum2 / n)

# calculate for standardized array; set up the standardized array value for futher calculation (Zij)
stand_array = my_array.copy()
stand_array.astype(float)
x_stand = 0
for i in range(0, nrow):
    for j in range(0, ncol):
        if my_array[i,j] != 0:
            x_stand = (float(my_array[i,j]) - x_ave) / x_std
        else:
            x_stand = 0
        stand_array[i,j] = (float(x_stand))
#set up the weights for the neighbor pixels
x_w = 0.125

#calculate Zscore_b2 from the standardized array.
# "Zscore_b2" means "b2i" in the E[Ii2] equation" 
# "Zscore_E" means "E[Ii]" in the Z score equation"
# "Zscore_A" means "A" in the E[Ii2] equation"; number 8 means the number of neighbours 
# "Zscore_B" means "B" in the E[Ii2] equation"; number 27 means 8!: the number of wik * wih
# "Zscore_V" means "V[Ii]" in the Z score equation"
Zscore_b2m4 = 0
Zscore_b2m2 = 0
for i in range(0, nrow):
    for j in range(0, ncol):
        if my_array[i,j] != 0:
            Zscore_b2m4 = Zscore_b2m4 + math.pow((float(my_array[i,j]) - x_ave), 4)
            Zscore_b2m2 = Zscore_b2m2 + math.pow((float(my_array[i,j]) - x_ave), 2)
Zscore_b2 = Zscore_b2m4 / math.pow(Zscore_b2m2, 2)

Zscore_E = 1 / (1 - float(n))
Zscore_A = (n - Zscore_b2) * math.pow(x_w, 2) * 8 / (n - 1)
Zscore_B = (2 * Zscore_b2 - n) * math.pow(x_w, 2) * 27 / ((n - 1) * (n - 2))
Zscore_V = Zscore_A - Zscore_B - math.pow(Zscore_E, 2) 


#copy the input raster as output raster
LMI_array = my_array.copy()
LMI_array.astype(float)
LMZ_array = my_array.copy()
LMZ_array.astype(float)
LMP_array = my_array.copy()
LMP_array.astype(float)
out_array = my_array.copy()
out_array.astype(float)

# Loop over the nump array to calculate Local moran's Index, Z score and P value  

LM_sum = 0
x_n = 0
x_pow = 0

for i in range(0, nrow):
    for j in range(0, ncol):
        if stand_array[i,j] != 0:
            for k in range(-1,2):
                for l in range(-1,2):
                    if (nrow > (i + k) >=0) and (ncol > (j + l) >=0) and stand_array[i+k,j+l] != 0:
                        LM_sum = LM_sum + float(stand_array[i+k,j+l]) * x_w
                        x_n = float(x_n) + 1
                        x_pow = x_pow + math.pow(float(stand_array[i+k,j+l]), 2)
            LM_index = (x_n - 2) * float(stand_array[i,j]) * (LM_sum - float(stand_array[i,j]) * x_w) / x_pow
            LM_Zscore = (LM_index - Zscore_E) / math.sqrt(Zscore_V)
            LM_Pvalue = st.norm.sf(abs(LM_Zscore))*2
        else:
            LM_index = 0
            LM_Zscore = 0
            LM_Pvalue = 0
        # select the standardized value < 0 & LM_index>0 & p value < 0.05 (coldspot clustered)
        if stand_array[i,j] < 0 and LM_index > 0 and LM_Pvalue <= 0.05:
            cold = 1
        else:
            cold = 2
        LMI_array[i,j] = (float(LM_index))
        LMZ_array[i,j] = (float(LM_Zscore))
        LMP_array[i,j] = (float(LM_Pvalue))
        out_array[i,j] = (int(cold))
        LM_sum = 0
        x_n = 0
        x_pow = 0

LMIRas = arcpy.NumPyArrayToRaster(LMI_array,lowerLeft,cellSize,value_to_nodata = 0)
LMIRas.save("LMIras")
LMZRas = arcpy.NumPyArrayToRaster(LMZ_array,lowerLeft,cellSize,value_to_nodata = 0)
LMZRas.save("LMZras")
LMPRas = arcpy.NumPyArrayToRaster(LMP_array,lowerLeft,cellSize,value_to_nodata = 0)
LMPRas.save("LMPras")
OutRas = arcpy.NumPyArrayToRaster(out_array,lowerLeft,cellSize,value_to_nodata = 0)
OutRas.save("Outras")
#get the coordinate system by describing a raster dataset
dsc = arcpy.Describe(inRaster)
coord_sys = dsc.spatialReference

arcpy.DefineProjection_management(LMIRas, coord_sys)
arcpy.DefineProjection_management(LMZRas, coord_sys)
arcpy.DefineProjection_management(LMPRas, coord_sys)
arcpy.DefineProjection_management(OutRas, coord_sys)

# # # control the mountain shadow
# # # OutRas1 = Con((B7B4Sel - OutRas == 0), 1 ,0)
# # # OutRas1.save("OutRasSel")

OutRas1 = Con((OutRas == 1), 1 ,0)
OutRas1.save("OutRasSel")

# # # # get the finished raster as output (only want to convert to arcGIS output)
# # # arcpy.gp.ExtractByMask_sa(LMIRas, clip_feature, LMIRaster)
# # # arcpy.gp.ExtractByMask_sa(LMZRas, clip_feature, LMZRaster)
# # # arcpy.gp.ExtractByMask_sa(LMPRas, clip_feature, LMPRaster)

#set up local polygon data as shape file
water_temp = "water_temp.shp"
water_area = "water_area.shp"
water_eli = "water_eli.shp"
water_sel = "water_sel.shp"
water_sim = "water_sim.shp"

# covert coldspot (output raster) to polygon
arcpy.RasterToPolygon_conversion(OutRas1, water_temp, "NO_SIMPLIFY", "VALUE")
# add the F_Area field
arcpy.CalculateAreas_stats(water_temp, water_area)
# eliminate small polygons (area smaller than 40000 square meters)
arcpy.EliminatePolygonPart_management(water_area, water_eli, "AREA", 40000, "", "ANY")
# select the lake part and area large than 100000 square meters
arcpy.Select_analysis(water_eli, water_sel, "\"gridcode\" = 1 AND \"F_Area\" > 100000")
# Simplify lake polygons.
arcpy.SimplifyPolygon_cartography(water_sel, water_sim, "POINT_REMOVE", 50, 100000, "RESOLVE_ERRORS", "KEEP_COLLAPSED_POINTS")
# smooth the polygon
arcpy.SmoothPolygon_cartography(water_sim, Lake, "PAEK", 100, "", "FLAG_ERRORS")
